/*
 * $RCSfile: crc32.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 * $Log: crc32.c,v $
 * Revision 1.1.1.1  1994/06/10 17:47:21  menze
 * Fixed a pointer type conflict
 *
 * Revision 1.1  1994/06/09  00:32:43  dcs
 * Initial revision
 */

/* The CRC protocol pushes a 16byte cryptographic checksum onto the front
   of a message.  On the incoming side, it pops a 16byte checksum, and
   compares it to the checksum computed for the rest of the message.
   The code for CRC is directly from the RSA distribution; it's a bit
   sluggish at 7.3 usec/byte, about twice as slow as DES. */


#include "xkernel.h"
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "crc_i.h"
#include "crchash.h"
#include "prot/hash.h" 
#include <krb5/krb5.h>
#include <krb5/crc-32.h>
#include <krb5/ext-proto.h>

extern int tracecrcp; int tracecrcp;
extern krb5_error_code crc32_sum_func();

/*  Byte order note:  All data is grouped in the order ABCD for CRC processing. */


#ifdef __STDC__

static XObj		crcCreateSessn( XObj, XObj, XObj, XObj);
static void		crcSessnInit( XObj );
static int		crcControlProtl( XObj, int, char *, int );
static xkern_return_t	crcDemux( XObj, XObj, Msg * );
static XObj		crcOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	crcOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	crcOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	crcOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	crcClose( XObj );
static xmsg_handle_t	crcPush( XObj, Msg * );
static xkern_return_t	crcPop( XObj, XObj, Msg *, void * );
static void 		crcStore( void *, char *, long, void * );
static int		crcControlSessn( XObj, int, char *, int );
static int		crcchunk( char *, long, CRC_CTX * );
static long 		crccopy(void *, char*, long, void * );
int  xk_crc(unsigned char [], unsigned char *, int, int, unsigned char []);
void crc_init( XObj );

#else

static XObj		crcCreateSessn();
static void		crcSessnInit();
static int		crcControlProtl();
static xkern_return_t	crcDemux();
static XObj		crcOpen();
static xkern_return_t	crcOpenEnable();
static xkern_return_t	crcOpenDisable();
static xkern_return_t	crcOpenDone();
static xkern_return_t	crcClose();
static xmsg_handle_t	crcPush();
static xkern_return_t	crcPop();
static int		crcControlSessn();
int  xk_crc();

#endif /* __STDC__ */

#undef xTrace0
#undef xTrace1
#undef xTrace2

#define xTrace0(a,b,c)     /* a b c */
#define xTrace1(a,b,c,d)   /* a b c d */
#define xTrace2(a,b,c,d,e) /* a b c d e */

void
crc_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace0(crcp, TR_ERRORS, "crc_init entered");
    xAssert(xIsProtocol(self));
    
    if ( ! xIsProtocol(llp = xGetDown(self, 0)) )
    { xError("crc_init: crc cannot get lower protocol object");	return; }

    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(CRC_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(CRC_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;

    self->control = crcControlProtl;
    self->open = crcOpen;
    self->openenable = crcOpenEnable;
    self->opendisable = crcOpenDisable;
    self->opendone = crcOpenDone;
    self->close = crcClose;  
    self->demux = crcDemux;
    xTrace0(crcp, TR_EVENTS, "crc_init exiting");
}



static XObj
crcOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	crcSessn;
    XObj	lls;
    PState	*pstate;

    pstate = (PState *) self->state;
    
    xTrace4(crcp,TR_EVENTS,"crcOpen calling xOpen with args %x %x %x %x",
	    self,hlpType,xGetDown(self,0),part);

    lls = xOpen(self, hlpType, xGetDown(self, 0), part);
    if ( lls == ERR_XOBJ )
    { xTrace0(crcp, TR_ERRORS, "crcOpen: could not open lls"); return ERR_XOBJ; }

    if ( mapResolve(pstate->actMap, &lls, &crcSessn) == XK_SUCCESS ) {
        xTrace0(crcp, TR_MAJOR_EVENTS, "found an existing one");
        xClose(lls);
    } else {
        xTrace0(crcp, TR_MAJOR_EVENTS, "creating a new one");
        if ( (crcSessn = crcCreateSessn(self, hlpRcv, hlpType, lls)) == ERR_XOBJ ) {
	  xTrace0(crcp, TR_ERRORS, "crcOpen: create session failed");
	  xClose(lls);
        }
    }

    return (crcSessn);
}


static XObj
crcCreateSessn( self, hlpRcv, hlpType, lls)
    XObj	self, hlpRcv, hlpType, lls;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    XObj        SessnArray[1];

    xTrace4(crcp, TR_EVENTS, "crcCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    ss = X_NEW(SState);

    SessnArray[0]=lls;
    s = xCreateSessn(crcSessnInit, hlpRcv, hlpType, self, 1, SessnArray);
    s->binding = mapBind(ps->actMap, &lls, s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(crcp, TR_ERRORS, "crcCreateSessn: mapBind error");
	xDestroy(s);
	xFree((char *) ss);
	return ERR_XOBJ;
      }

    xTrace1(crcp, TR_FULL_TRACE, "crcCreateSessn: X_NEW(SState) returned %x", ss);

    s->state = (VOID *)ss;
    ss->flags = 0;
    ss->LprefixL = 0; ss->Lprefix = (unsigned char *)0;
    ss->RprefixL = 0; ss->Rprefix = (unsigned char *)0;
    xTrace1(crcp, TR_MORE_EVENTS, "crcCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
crcOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(crcp, TR_EVENTS, "crcOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(crcp, TR_MORE_EVENTS, "crcOpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }
    
    val = xOpenEnable(self, hlpType, xGetDown(self,0), part);
    if ( val == XK_FAILURE ) {
        xTrace0(crcp, TR_MORE_EVENTS, "crcOpenEnable xOpenEnable failed ");
        return XK_FAILURE;
    }
    
    xTrace1(crcp, TR_MORE_EVENTS, "crcOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
crcOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(crcp, TR_EVENTS, "crcOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(crcp, TR_MORE_EVENTS, "crcOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
crcOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
   
    xTrace0(crcp, TR_MAJOR_EVENTS, "Entered crcOpenDone");

    if ( self == hlpType ) {
        xTrace0(crcp, TR_ERRORS, "self == hlpType in crcOpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(crcp, TR_ERRORS,
                "crcOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    if ( (s = crcCreateSessn(self, e->hlpRcv, e->hlpType, lls))
                == ERR_XOBJ ) {
        xClose(lls);
        return XK_FAILURE;
    }

    xTrace0(crcp, TR_EVENTS,
            "crcOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}



static xkern_return_t
crcDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val;
  
  xTrace3(crcp, TR_EVENTS, "crcDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(crcp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(crcp, TR_EVENTS, "crcDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);

  if (val == XK_SUCCESS)
  xTrace0(crcp, TR_FUNCTIONAL_TRACE, "crcDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(crcp, TR_FUNCTIONAL_TRACE, "crcDemux returning XK_FAILURE");
  else
  xTrace1(crcp, TR_FUNCTIONAL_TRACE, "crcDemux returning %x", val);

  return val;
}


static xkern_return_t
crcClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    SState	*ss;
    xkern_return_t res;

    xTrace1(crcp, TR_MAJOR_EVENTS, "crcClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    ss = (SState *)(s->state);
    if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
      /* erase local prefix, in case it's sensitive */
    { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL);
      xFree((VOID *)ss->Lprefix); }
    ss->LprefixL = 0; ss->Lprefix = 0;
    if (ss->Rprefix)  /* erase remote prefix, in case it's sensitive */
    { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL);
      xFree((VOID *)ss->Rprefix); }
    ss->RprefixL = 0; ss->Rprefix = 0;
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    xClose(xGetDown(s,0));
    xDestroy( s );
    xTrace0(crcp, TR_DETAILED, "crcClose returning");
    return XK_SUCCESS;
}

static
long crccopy(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
    xTrace4(crcp, TR_FULL_TRACE, "crccopy args: %x %x %x %x",
	    to,from,len,ignore);

    bcopy(from,to,len);
    return len;
}

static
void crcStore(src, dst, len, arg)
    void *src;
    char *dst;
    long len;
    void *arg;
{
    xTrace4(crcp, TR_FULL_TRACE, "crcStore args: %x %x %x %x",
	    src,dst,len,arg);

    bcopy((char *)src,dst,len);
}


static int
crcchunk( msgptr, msglen, crcctxp )
  char *msgptr; long msglen; CRC_CTX *crcctxp;
{ 
#if 0
CRCUpdate(crcctxp,(unsigned char *)msgptr,(unsigned int)msglen); 
#endif
return TRUE; 
}


static xmsg_handle_t
crcPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    long i;
    /* CRC_CTX crcctx; */
    SState *ss;
    krb5_checksum  the_sum;
    int len, clen, diff;
    char *flat_msg;
    krb5_error_code retval;
    

    xTrace2(crcp, TR_EVENTS, "crcPush: session %x  msg %x", s, msg);
    xAssert(xIsSession(s));
    ss = (SState *)(s->state);

    the_sum.contents = (u_char *)xMalloc(4);
    len = msgLen(msg);
    clen = ((len + 7) / 8) * 8;
    diff = clen - len;

    flat_msg = (char *)xMalloc(clen);
    msgPop(msg, crccopy, flat_msg, len, 0);
    if (diff) {
      memset(flat_msg+len, 0, diff);
    }
    
    retval = crc32_sum_func(flat_msg, clen, 0, 0, &the_sum);
    if (retval) {
      xTrace1(crcp, TR_ERRORS, "error (%d) while checksumming", retval);
    }
    ((int*)flat_msg)[2] = *((int *)the_sum.contents);
    msgPush(msg, crcStore, flat_msg, clen, 0);
    /* msgPush(msg, crcStore, the_sum.contents, 4, 0); */

#if 0
     CRCInit(&crcctx); 
    if (ss->LprefixL && ss->Lprefix) CRCUpdate(&crcctx,ss->Lprefix,ss->LprefixL);
    msgForEach(msg,(XCharFun)crcchunk,&crcctx); /* Feed each msg segment to CRCUpdate. */
    CRCFinal(&crcctx);
    msgPush(msg,crcStore,&crcctx.digest[0],16,0);  /* Push digest onto msg. */

    xIfTrace (crcp, TR_MORE_EVENTS)
    {  printf("crcPush: message length %d; digest is ", msgLen(msg));
       for (i=0;i<16;i++) printf(" %02x", crcctx.digest[i]);
       printf("\n"); }
#endif

    return xPush(xGetDown(s, 0), msg);
}



static xkern_return_t
crcPop( s, lls, msg, hdr)
    XObj	s, lls;
    Msg 	*msg;
    void        *hdr;
{
    long len; int i;
    CRC_CTX crcctx;
    u_char digest[4]; 

    SState *ss;

    xTrace4(crcp, TR_EVENTS, "crcPop: args %x %x %x %x", s, lls, msg, hdr);
#if 0
    ss = (SState *)(s->state);

    len = msgLen(msg);
    if (len<4)
    { xTrace1(crcp, TR_ERRORS,
	      "crcPop received bad message, length %D less than 16",len);
      return XK_FAILURE; }

    msgPop(msg, crccopy, digest, 4, 0);  /* read message digest */

    CRCInit(&crcctx);
    if (ss->RprefixL && ss->Rprefix) CRCUpdate(&crcctx,ss->Rprefix,ss->RprefixL);
    msgForEach(msg,(XCharFun)crcchunk,&crcctx); /* Feed each msg segment to CRCUpdate. */
    CRCFinal(&crcctx);

    xIfTrace (crcp, TR_MORE_EVENTS)
    {  printf("crcPop: message length 16+%d; digest is ", len-16);
       for (i=0;i<16;i++) printf(" %02x", crcctx.digest[i]);
       printf("\n"); }

    for (i=0;i<16;i++)  /* compare received & computed digests */
    {  if ( digest[i] != crcctx.digest[i] )
       {  xTrace0(crcp, TR_ERRORS, "crcPop: bad compare for message digest");
	  xIfTrace (crcp, TR_EVENTS)
	  {  printf("crcPop: digest from header is ");
	     for (i=0;i<16;i++) printf(" %02x", digest[i]);
	     printf("\ncrcPop: computed digest is    ");
	     for (i=0;i<16;i++) printf(" %02x", crcctx.digest[i]);
	     printf("\n"); }
	  break; }; }

    /* Fixup length in IP pseudoheader? */
    if (((SState *)(s->state))->flags & 1 || ((PState *)(xMyProtl(s)->state))->flags & 1)
    {  IPpseudoHdr *ipphdr = msgGetAttr( msg, 0);
       if (ipphdr)
       {  ipphdr->len = htons(ntohs(ipphdr->len)-16);
	  xTrace3(crcp, TR_FULL_TRACE,
	    "crcPop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len before %d",
		  ipphdr, ntohs(ipphdr->len), len);  }
       else { xTrace0(crcp, TR_FULL_TRACE,
	       "crcPop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); };
     }
#endif
    return xDemux(s, msg);    
} 


static int
crcControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(s, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= 16; if (max<0) return 0;
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case HASH_SETPREFIX:
	{  SState *ss = (SState *)(s->state);
	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (ss->Rprefix)  /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->LprefixL = ss->RprefixL = len;
	      ss->Lprefix =  ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len); }
	   else { ss->LprefixL =  ss->RprefixL = 0; ss->Lprefix = ss->Rprefix = 0; }
	   return 0;  }

      case HASH_SETLOCALPREFIX:
	{  SState *ss = (SState *)(s->state);
	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (len > 0)
	   {  ss->LprefixL = len; ss->Lprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len);  }
	   else { ss->LprefixL = 0; ss->Lprefix = 0; }
	   return 0;  }

      case HASH_SETREMOTEPREFIX:
	{  SState *ss = (SState *)(s->state);
	   if (ss->Rprefix && ss->Rprefix != ss->Lprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->RprefixL = len; ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Rprefix,len);  }
	   else { ss->RprefixL = 0; ss->Rprefix = 0; }
	   return 0;  }

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= 1;
	    xTrace0(crcp, TR_EVENTS,
	       "crcControlSessn: Turning on IPpseudoheader length fixup.");  }
      /* **** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too. **** */

      default:
	return xControl(xGetDown(s,0), op, buf, len);
    }
}


static int
crcControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsProtocol(self));

    switch (op) {
      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(self, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= 16; if (max<0) return 0;
            *(int *)buf = max;
            return(sizeof(max));
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PState *)(self->state))->flags |= 1;
	    xTrace0(crcp, TR_EVENTS,
	       "crcControlProtl: Turning on IPpseudoheader length fixup.");  }
      /* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too. **** */

      default:
	return xControl(xGetDown(self, 0), op, buf, len);
  }
  
}


static void
crcSessnInit(s)
    XObj s;
{
  s->push = crcPush;
  s->pop = crcPop;
  s->close = crcClose;
  s->control = crcControlSessn;
}

